/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"

#include <iterator>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <Eigen/src/Geometry/Transform.h>

#include "ArmarXCore/core/services/tasks/TaskUtil.h"

#include "RobotAPI/libraries/core/remoterobot/RemoteRobot.h"

// Navigation
#include "Navigation/libraries/core/types.h"
#include "Navigation/libraries/util/util.h"
#include <Navigation/libraries/factories/NavigationStackFactory.h>


armarx::nav::components::Navigator::Navigator() : executor{platformUnit}
{
}

armarx::nav::components::Navigator::~Navigator()
{
}


void
armarx::nav::components::Navigator::onInitComponent()
{
}


void
armarx::nav::components::Navigator::onConnectComponent()
{
}


void
armarx::nav::components::Navigator::onDisconnectComponent()
{
}


void
armarx::nav::components::Navigator::onExitComponent()
{
}


void
armarx::nav::components::Navigator::updateContext()
{
    scene.staticScene = staticScene();
    scene.robot = getRobot();
}


std::string
armarx::nav::components::Navigator::getDefaultName() const
{
    return "Navigator";
}


std::vector<armarx::nav::core::Pose>
convert(const std::vector<Eigen::Matrix4f>& wps)
{
    using namespace armarx::nav;
    std::vector<core::Pose> p;
    p.reserve(wps.size());
    std::transform(wps.begin(), wps.end(), std::back_inserter(p), [](const auto & p)
    {
        return Eigen::Affine3f(p);
    });
    return p;
}


std::string
armarx::nav::components::Navigator::createConfig(const aron::data::AronDictPtr& stackConfig,
        const Ice::Current&)
{
    // TODO: Not thread-safe.
    const std::string uuid = boost::uuids::to_string(boost::uuids::random_generator()());

    server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);
    navigators.emplace(uuid, server::Navigator{stack, scene, executor});

    return uuid;
}


void
armarx::nav::components::Navigator::moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
        const std::string& navigationMode,
        const std::string& configId,
        const Ice::Current&)
{
    // TODO: Error handling.

    navigators.at(configId).moveTo(convert(waypoints), core::NavigationFramesMap.from_name(navigationMode));
}


void
armarx::nav::components::Navigator::moveTowards(const Eigen::Vector3f& direction,
        const std::string& navigationMode,
        const std::string& configId,
        const Ice::Current&)
{
    // TODO: Error handling.

    navigators.at(configId).moveTowards(direction, core::NavigationFramesMap.from_name(navigationMode));
}


void
armarx::nav::components::Navigator::pauseMovement(const Ice::Current&)
{

}


void
armarx::nav::components::Navigator::resumeMovement(const Ice::Current&)
{

}


armarx::PropertyDefinitionsPtr
armarx::nav::components::Navigator::createPropertyDefinitions()
{
    armarx::PropertyDefinitionsPtr def =
        new ComponentPropertyDefinitions(getConfigIdentifier());

    // Publish to a topic (passing the TopicListenerPrx).
    // def->topic(myTopicListener);

    // Subscribe to a topic (passing the topic name).
    // def->topic<PlatformUnitListener>("MyTopic");

    // Use (and depend on) another component (passing the ComponentInterfacePrx).
    def->component(platformUnit);

    // Add a required property.
    // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

    // Add an optionalproperty.
    // def->optional(properties.numBoxes, "p.box.Number", "Number of boxes to draw in ArViz.");

    return def;
}


armarx::nav::core::StaticScene
armarx::nav::components::Navigator::staticScene()
{
    core::StaticScene scene;

    objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();

    scene.objects = util::asSceneObjects(objectPoses);

    return scene;
}


VirtualRobot::RobotPtr
armarx::nav::components::Navigator::getRobot()
{
    auto robot = RemoteRobot::createLocalCloneFromFile(
                     getRobotStateComponent(), VirtualRobot::RobotIO::RobotDescription::eFull);
    // auto robot = RemoteRobot::createLocalClone(getRobotStateComponent());
    ARMARX_CHECK_NOT_NULL(robot);

    // ARMARX_INFO << "Resetting robot";
    // robot = VirtualRobot::RobotIO::loadRobot(robot->getFilename());

    RemoteRobot::synchronizeLocalClone(robot, getRobotStateComponent());

    return robot;
}
